"""This script reads from the webcam with OpenCV and publishes the images to a topic over shared memory.
Another process, started from this same script, subscribes to the images and shows them in a window.

This script requires that you have OpenCV installed, you can do this with: `pip install opencv-contrib-python`.
"""
from dataclasses import dataclass
from typing import Final

import cv2
import numpy as np
from cyclonedds.idl import IdlStruct
from loguru import logger

from airo_ipc.cyclone_shm.idl_shared_memory.base_idl import BaseIdl
from airo_ipc.framework.framework import IpcKind
from airo_ipc.framework.node import Node

TOPIC_RESOLUTION: Final[str] = "resolution"
TOPIC_BGR: Final[str] = "frame"


@dataclass
class ResolutionIdl(IdlStruct):
    """We will send the resolution of the webcam over DDS: we need to define an IDL struct for this."""
    width: int
    height: int


@dataclass
class WebcamFrame(BaseIdl):
    """We will send the BGR frames over shared memory: we need to derive from BaseIDL."""
    bgr: np.ndarray

    @staticmethod
    def with_resolution(width: int, height: int):
        """We may not know the resolution of the webcam when we create the frame, so we need a factory method."""
        return WebcamFrame(bgr=np.zeros((height, width, 3), dtype=np.uint8))


class WebcamPublisher(Node):
    """The publisher will open the webcam and publish the resolution and frame in a loop."""
    def _setup(self):
        logger.info("Opening webcam.")
        self._camera = cv2.VideoCapture(0)

        logger.info("Getting resolution.")
        width, height = int(self._camera.get(cv2.CAP_PROP_FRAME_WIDTH)), int(
            self._camera.get(cv2.CAP_PROP_FRAME_HEIGHT))

        logger.info("Registering publishers.")
        self._register_publisher(TOPIC_RESOLUTION, ResolutionIdl, IpcKind.DDS)
        self._register_publisher(TOPIC_BGR, WebcamFrame.with_resolution(width, height), IpcKind.SHARED_MEMORY)

    def _step(self):
        """The _step method is called in a loop by the Node superclass."""
        ret, frame = self._camera.read()

        if not ret:
            logger.error("Could not read frame from webcam. Stopping publisher.")
            self.stop()

        self._publish(TOPIC_RESOLUTION, ResolutionIdl(width=frame.shape[1], height=frame.shape[0]))
        self._publish(TOPIC_BGR, WebcamFrame(bgr=frame))

    def _teardown(self):
        """This is where we can clean up resources when the publisher is stopped."""
        self._camera.release()


class WebcamSubscriber(Node):
    """The subscriber will show the frames in a window."""
    def _setup(self):
        logger.info("Creating webcam window.")
        cv2.namedWindow("Webcam", cv2.WINDOW_NORMAL)
        logger.info("Subscribing to resolution messages.")
        self._subscribe(TOPIC_RESOLUTION, ResolutionIdl, IpcKind.DDS, self._on_receive_resolution)
        # We cannot yet subscribe to the BGR topic, because we do not yet know the resolution. We will do this in the
        # resolution callback.

    def _on_receive_resolution(self, resolution: ResolutionIdl):
        # If we already subscribed to the BGR topic, we do not need to do it again.
        if TOPIC_BGR in self._readers:
            return

        # It is safe to subscribe to topics in callbacks, because a copy of the existing subscriptions is made in the
        # Node superclass.
        logger.info("Received resolution message. Subscribing to BGR messages.")
        self._subscribe(TOPIC_BGR, WebcamFrame.with_resolution(resolution.width, resolution.height),
                        IpcKind.SHARED_MEMORY, self._on_receive_frame)

    def _on_receive_frame(self, frame: WebcamFrame):
        # Here, we show the frame in a window in the callback function. This is not ideal, because the callback function
        # should be as fast as possible. In a real application, you could use a queue to pass the frames to another
        # thread that shows the frames in a window.
        cv2.imshow("Webcam", frame.bgr)
        key = cv2.waitKey(1)
        if key == ord("q"):
            logger.info("Closing webcam window.")
            self.stop()

    def _step(self):
        pass

    def _teardown(self):
        # We need to close the window when the subscriber is stopped.
        cv2.destroyAllWindows()


if __name__ == '__main__':
    logger.info("Creating publisher.")
    publisher = WebcamPublisher(20, True)
    logger.info("Starting publisher.")
    publisher.start()

    logger.info("Creating subscriber.")
    subscriber = WebcamSubscriber(20, True)
    logger.info("Starting subscriber.")
    subscriber.start()

    logger.info("Joining subscriber: will quit when the user pressed 'q' with the CV2 window in focus.")
    subscriber.join()
    publisher.stop()



from typing import Any
from cyclonedds.domain import DomainParticipant
from cyclonedds.idl import IdlMeta
from cyclonedds.sub import DataReader
from cyclonedds.topic import Topic

from airo_ipc.cyclone_shm.defaults import CYCLONE_DEFAULTS


class DDSReader:
    def __init__(
        self,
        domain_participant: DomainParticipant,
        topic_name: str,
        idl_dataclass: IdlMeta,
    ):
        self.topic: Topic = Topic(domain_participant, topic_name, idl_dataclass)
        self.reader = DataReader(domain_participant, self.topic, CYCLONE_DEFAULTS.QOS)

    def __call__(self) -> Any:
        """Get the latest data from the reader.

        Returns:
            An instance of the idl_dataclass containing the data, or `None`.
        """
        data = self.reader.read()
        if len(data) == 0:
            return None
        return data[-1]

    def take(self) -> Any:
        """Take the latest data from the reader.

        Only returns samples that have not yet been returned in a take operation.
        (https://cyclonedds.io/docs/cyclonedds/latest/about_dds/datareaders.html)

        Returns:
            An instance of the idl_dataclass containing the data, or `None`."""
        data = self.reader.take()
        if len(data) == 0:
            return None
        return data[-1]



from cyclonedds.domain import DomainParticipant
from cyclonedds.idl import IdlMeta, IdlStruct
from cyclonedds.pub import DataWriter
from cyclonedds.topic import Topic

from airo_ipc.cyclone_shm.defaults import CYCLONE_DEFAULTS


class DDSWriter:
    def __init__(
        self,
        domain_participant: DomainParticipant,
        topic_name: str,
        idl_dataclass: IdlMeta,
    ):
        self.topic: Topic = Topic(domain_participant, topic_name, idl_dataclass)
        self.writer = DataWriter(domain_participant, self.topic, CYCLONE_DEFAULTS.QOS)

    def __call__(self, msg: IdlStruct) -> None:
        self.writer.write(msg)



"""
This module defines the BufferTemplate class, which serves as a base class for buffer templates using dataclasses.

Classes:
    BufferTemplate: Base class that provides a method to retrieve numpy array field information.
"""

from dataclasses import dataclass, fields
from typing import Any, Generator, Iterator, Tuple
from typing_extensions import deprecated

import numpy as np


@dataclass
class BaseIdl:
    """
    Base class for buffer templates using dataclasses.

    Provides a method to retrieve information about numpy array fields,
    which is useful for setting up shared memory buffers for inter-process communication.
    """

    def get_fields(self) -> Iterator[Tuple[str, Tuple[int, ...], Any, int]]:
        """
        Generator that yields information about the numpy array fields.

        Iterates over all fields defined in the dataclass and yields a tuple
        containing the field name, shape, dtype, and number of bytes for each numpy array field.

        Yields:
            tuple: (field_name, shape, dtype, nbytes) of each numpy array field.
        """
        for field in fields(self):
            array = getattr(self, field.name)
            # Only process fields that are numpy arrays
            if isinstance(array, np.ndarray):
                yield (field.name, array.shape, array.dtype, array.nbytes)
            else:
                raise AttributeError(
                    f"BaseIDL subclasses only support NumPy types: {field.name} violates this."
                )


@deprecated("Use BaseIdl instead.")
class BaseIDL(BaseIdl):
    """Present for backwards compatibility, use BaseIdl instead."""



import multiprocessing
import time
from abc import abstractmethod, ABC
from multiprocessing.context import SpawnProcess
from typing import Dict, Union, Callable, Any

from cyclonedds.domain import DomainParticipant
from cyclonedds.idl import IdlMeta
from loguru import logger

from airo_ipc.cyclone_shm.idl_shared_memory.base_idl import BaseIdl
from airo_ipc.cyclone_shm.patterns.ddsreader import DDSReader
from airo_ipc.cyclone_shm.patterns.ddswriter import DDSWriter
from airo_ipc.cyclone_shm.patterns.sm_reader import SMReader
from airo_ipc.cyclone_shm.patterns.sm_writer import SMWriter
from airo_ipc.framework.framework import IpcKind


class Node(SpawnProcess, ABC):
    """A Node is a process that can communicate with other Nodes using the IPC framework.
    This is a convenience class: it is not required to use airo-ipc, but it can simplify some of your code,
    abstracting away implementation details and providing a simple publish/subscribe system."""

    def __init__(self, update_frequency: float, verbose: bool = False):
        super().__init__()

        self._update_frequency = update_frequency
        self._stop_event = multiprocessing.get_context("spawn").Event()
        self._verbose = verbose

    def run(self) -> None:
        self._node_setup()

        try:
            while not self._stop_event.is_set():
                start_time = time.time()

                self._update_subscriptions()

                self._step()

                elapsed_time = time.time() - start_time
                desired_sleep_time = (1.0 / self._update_frequency) - elapsed_time
                sleep_time = max(0.0, desired_sleep_time)

                if self._verbose and desired_sleep_time < 0.0:
                    logger.warning(
                        f"Node {self.__class__.__name__} cannot keep up with desired update frequency {self._update_frequency}Hz!"
                    )

                time.sleep(sleep_time)

            self._teardown()
        finally:
            # SM readers/writers own SHM segments that must be released before the
            # DomainParticipant is closed (they publish/subscribe to a buffer-nr topic
            # owned by the participant).
            for reader in self._readers.values():
                stop = getattr(reader, "stop", None)
                if callable(stop):
                    try:
                        stop()
                    except Exception:
                        logger.exception("Error stopping reader")
            for writer in self._writers.values():
                stop = getattr(writer, "stop", None)
                if callable(stop):
                    try:
                        stop()
                    except Exception:
                        logger.exception("Error stopping writer")
            self._readers.clear()
            self._writers.clear()

            close = getattr(self._cyclone_dp, "close", None)
            if callable(close):
                try:
                    close()
                except Exception:
                    logger.exception("Error closing DomainParticipant")

    def _node_setup(self) -> None:
        """Handles setup that must be done inside the child process."""
        self._cyclone_dp = DomainParticipant()

        self._readers: Dict[str, Union[SMReader, DDSReader]] = dict()
        self._callbacks: Dict[str, Callable] = dict()

        self._writers: Dict[str, Union[SMWriter, DDSWriter]] = dict()

        self._setup()

    def _subscribe(
        self,
        topic_name: str,
        idl_dataclass: Union[IdlMeta, BaseIdl],
        ipc_kind: IpcKind,
        callback: Callable,
    ) -> None:
        if topic_name in self._readers:
            if self._verbose:
                logger.warning(
                    f"Node {self.__class__.__name__} is already subscribed to topic {topic_name}. Ignoring."
                )
            return

        if ipc_kind == IpcKind.DDS:
            assert isinstance(idl_dataclass, IdlMeta)
            self._readers[topic_name] = DDSReader(
                self._cyclone_dp, topic_name, idl_dataclass
            )
        elif ipc_kind == IpcKind.SHARED_MEMORY:
            assert isinstance(idl_dataclass, BaseIdl)
            self._readers[topic_name] = SMReader(
                self._cyclone_dp, topic_name, idl_dataclass
            )
        else:
            raise NotImplementedError(
                f"No implementation available for subscribing with IPC kind {ipc_kind}."
            )

        self._callbacks[topic_name] = callback

    def _update_subscriptions(self) -> None:
        # When a subscription is made in a subscription callback, we should not update it in the same iteration.
        # So we take a snapshot of the items and iterate over that.
        items = list(self._readers.items())
        for topic_name, reader in items:
            value = reader()
            if value is not None:
                self._callbacks[topic_name](value)

    def _register_publisher(
        self, topic_name: str, idl_dataclass: Union[IdlMeta, BaseIdl], ipc_kind: IpcKind
    ) -> None:
        if topic_name in self._writers:
            if self._verbose:
                logger.warning(
                    f"Node {self.__class__.__name__} is already registered as a publisher for topic {topic_name}. Ignoring."
                )
            return

        if ipc_kind == IpcKind.DDS:
            assert isinstance(idl_dataclass, IdlMeta)
            self._writers[topic_name] = DDSWriter(
                self._cyclone_dp, topic_name, idl_dataclass
            )
        elif ipc_kind == IpcKind.SHARED_MEMORY:
            assert isinstance(idl_dataclass, BaseIdl)
            self._writers[topic_name] = SMWriter(
                self._cyclone_dp, topic_name, idl_dataclass
            )
        else:
            raise NotImplementedError(
                f"No implementation available for publishing with IPC kind {ipc_kind}."
            )

    def _publish(self, topic_name: str, value: Any) -> None:
        if topic_name not in self._writers:
            logger.error(
                f"Node {self.__class__.__name__} does not have a writer for topic {topic_name}."
            )
            raise RuntimeError(
                f"Cannot publish topics that were not registered. Violating topic: {topic_name}."
            )

        self._writers[topic_name](value)

    @abstractmethod
    def _setup(self) -> None:
        pass

    @abstractmethod
    def _step(self) -> None:
        pass

    @abstractmethod
    def _teardown(self) -> None:
        pass

    def stop(self) -> None:
        self._stop_event.set()



from cyclonedds.qos import Qos, Policy
from cyclonedds.util import duration


class CYCLONE_DEFAULTS:
    QOS = Qos(
        Policy.Reliability.BestEffort,
        Policy.History.KeepLast(1),
    )
    QOS_RPC = Qos(
        Policy.Reliability.Reliable(max_blocking_time=60),
        Policy.Deadline(duration(milliseconds=1000)),
        Policy.History.KeepLast(1),
        Policy.ResourceLimits(
            max_samples=1, max_instances=1, max_samples_per_instance=1
        ),
    )
    QOS_RPC_STATUS = Qos(
        Policy.Reliability.Reliable(max_blocking_time=1),
        Policy.Deadline(duration(milliseconds=1000)),
        Policy.History.KeepLast(1),
        Policy.ResourceLimits(
            max_samples=1, max_instances=1, max_samples_per_instance=1
        ),
    )



"""This script implements the airo-camera-toolkit RGBCamera interface for a webcam using OpenCV.

This script reads from the webcam with OpenCV and publishes the images to a topic over shared memory.
Another process, started from this same script, subscribes to the images and shows them in a window.

This script requires you to install the airo-camera-toolkit, which you can do by following the instructions here:
https://github.com/airo-ugent/airo-mono
"""
import time
from dataclasses import dataclass
from typing import Final

import cv2
import numpy as np
from airo_camera_toolkit.cameras.opencv_videocapture.opencv_videocapture import OpenCVVideoCapture
from airo_camera_toolkit.interfaces import RGBCamera
from airo_camera_toolkit.utils.image_converter import ImageConverter
from airo_typing import CameraIntrinsicsMatrixType, NumpyIntImageType, NumpyFloatImageType, CameraResolutionType
from cyclonedds.domain import DomainParticipant
from cyclonedds.idl import IdlStruct
from loguru import logger

from airo_ipc.cyclone_shm.idl_shared_memory.base_idl import BaseIdl
from airo_ipc.cyclone_shm.patterns.ddsreader import DDSReader
from airo_ipc.cyclone_shm.patterns.sm_reader import SMReader
from airo_ipc.framework.framework import IpcKind
from airo_ipc.framework.node import Node

TOPIC_RESOLUTION: Final[str] = "resolution"
TOPIC_RGB: Final[str] = "frame"


@dataclass
class ResolutionIdl(IdlStruct):
    """We will send the resolution of the webcam over DDS: we need to define an IDL struct for this."""
    width: int
    height: int


@dataclass
class WebcamFrame(BaseIdl):
    """We will send the RGB frames over shared memory: we need to derive from BaseIDL."""
    rgb: np.ndarray
    intrinsics: np.ndarray

    @staticmethod
    def with_resolution(width: int, height: int):
        """We may not know the resolution of the webcam when we create the frame, so we need a factory method."""
        return WebcamFrame(rgb=np.zeros((height, width, 3), dtype=np.uint8), intrinsics=np.zeros((3, 3)))


class RGBCameraPublisher(Node):
    """The publisher will open the webcam and publish the resolution and frame in a loop."""

    def _setup(self):
        logger.info("Opening camera.")
        self._camera = OpenCVVideoCapture(intrinsics_matrix=np.eye(3))

        logger.info("Getting resolution.")
        width, height = self._camera.resolution

        logger.info("Registering publishers.")
        self._register_publisher(TOPIC_RESOLUTION, ResolutionIdl, IpcKind.DDS)
        self._register_publisher(TOPIC_RGB, WebcamFrame.with_resolution(width, height), IpcKind.SHARED_MEMORY)

    def _step(self):
        """The _step method is called in a loop by the Node superclass."""
        rgb = self._camera.get_rgb_image_as_int()

        self._publish(TOPIC_RESOLUTION,
                      ResolutionIdl(width=self._camera.resolution[0], height=self._camera.resolution[1]))
        self._publish(TOPIC_RGB, WebcamFrame(rgb=rgb, intrinsics=self._camera.intrinsics_matrix()))

    def _teardown(self):
        pass


class WebcamSubscriber(RGBCamera):
    """Remark how we don't *need* to inherit from the Node superclass; this only depends on your use cases.
        Here, we can simplify our code by not using callbacks, but running the subscriber in the main process."""

    def __init__(self):
        super().__init__()

        self._cyclone_dp = DomainParticipant()
        self._reader_resolution = DDSReader(self._cyclone_dp, TOPIC_RESOLUTION, ResolutionIdl)
        # Wait for the first resolution message.
        resolution = None
        while resolution is None:
            resolution = self._reader_resolution()
            logger.info("Did not yet receive resolution message. Sleeping for 100 milliseconds...")
            time.sleep(0.1)
        self._reader_rgb = SMReader(self._cyclone_dp, TOPIC_RGB,
                                    WebcamFrame.with_resolution(resolution.width, resolution.height))

    @property
    def resolution(self) -> CameraResolutionType:
        return self._resolution

    def _retrieve_rgb_image(self) -> NumpyFloatImageType:
        return ImageConverter.from_numpy_int_format(self._rgb).image_in_numpy_format

    def _retrieve_rgb_image_as_int(self) -> NumpyIntImageType:
        return self._rgb

    def intrinsics_matrix(self) -> CameraIntrinsicsMatrixType:
        return self._intrinsics_matrix

    def _grab_images(self) -> None:
        frame = self._reader_rgb()
        if frame is not None:
            self._rgb = frame.rgb.copy() # SMReader does not copy arrays: we need to do this ourselves.
            self._intrinsics_matrix = frame.intrinsics


if __name__ == '__main__':
    logger.info("Creating publisher.")
    publisher = RGBCameraPublisher(20, True)
    logger.info("Starting publisher.")
    publisher.start()

    logger.info("Creating subscriber.")
    subscriber = WebcamSubscriber()

    cv2.namedWindow("Webcam", cv2.WINDOW_NORMAL)

    while True:
        rgb = subscriber.get_rgb_image_as_int()
        rgb_cv = ImageConverter.from_numpy_int_format(rgb).image_in_opencv_format

        cv2.imshow("Webcam", rgb_cv)
        key = cv2.waitKey(1)
        if key == ord("q"):
            logger.info("Stopping...")
            break

    publisher.stop()
    cv2.destroyAllWindows()



"""
This module provides classes to write data to shared memory using Cyclone DDS.

Classes:
    SMBufferWriteField: Manages a numpy array backed by shared memory for writing.
    SMWriter: Writes shared memory buffers using DDS for synchronization.
"""

import atexit
import time
from multiprocessing import shared_memory
from typing import Any, Dict, List, Tuple

import numpy as np
from cyclonedds.domain import DomainParticipant

from airo_ipc.cyclone_shm.idl.defaults.buffer_nr import BufferNrSample
from airo_ipc.cyclone_shm.idl_shared_memory.base_idl import BaseIdl
from airo_ipc.cyclone_shm.patterns.ddswriter import DDSWriter


class SMBufferWriteField:
    """
    Manages a numpy array backed by shared memory for writing.

    This class creates a shared memory segment and a numpy array
    that uses the shared memory as its buffer for writing data.
    """

    def __init__(self, name: str, shape: Tuple[int, ...], dtype: Any, nbytes: int):
        """
        Initialize the shared memory buffer field.

        Args:
            name (str): The name of the shared memory block.
            shape (tuple): The shape of the numpy array.
            dtype (numpy.dtype): The data type of the numpy array.
            nbytes (int): The number of bytes in the shared memory buffer.
        """

        try:
            self.shm = shared_memory.SharedMemory(create=True, size=nbytes, name=name)
        except FileExistsError:
            # this can occur if the process was killed and the shared memory was not cleaned up
            print(f"Shared memory file {name} exists. Will unlink and re-create it.")

            shm_old = shared_memory.SharedMemory(create=False, size=nbytes, name=name)
            shm_old.unlink()
            self.shm = shared_memory.SharedMemory(create=True, size=nbytes, name=name)
        # Create a new shared memory block with the given name and size
        # Create a numpy array that uses the shared memory buffer
        self.shared_array: np.ndarray = np.ndarray(
            shape, dtype=dtype, buffer=self.shm.buf
        )

    def stop(self) -> None:
        """Close and unlink the shared memory segment."""
        self.shm.close()
        self.shm.unlink()


class SMWriter:
    """
    Writes shared memory buffers using Cyclone DDS for synchronization.

    This class writes data to shared memory and publishes buffer numbers
    using DDS to synchronize readers.
    """

    def __init__(
        self,
        domain_participant: DomainParticipant,
        topic_name: str,
        idl_dataclass: BaseIdl,
        nr_of_buffers: int = 3,
    ):
        """
        Initialize the shared memory writer.

        Args:
            domain_participant (DomainParticipant): The DDS domain participant.
            topic_name (str): The base name of the topic.
            idl_dataclass (BaseIdl): The template defining the buffer structure.
            nr_of_buffers (int): The number of shared memory buffers per field. Raise this value to lower the chance of data races. Lower values reduce memory usage at the risk of data races. This value should be synchronized between writers and readers.
        """
        self.domain_participant = domain_participant
        self.topic_name = topic_name
        self.buffer_template = idl_dataclass
        self.nr_of_buffers = nr_of_buffers

        # Create a DDS writer for buffer numbers
        self.buffer_nr_writer = DDSWriter(
            domain_participant=domain_participant,
            topic_name=f"{topic_name}__buffer_nr",
            idl_dataclass=BufferNrSample,
        )

        # Create shared memory buffers
        self.buffers: List[Dict[str, SMBufferWriteField]] = self.__make_shared_memory()
        self.buffer_idx = 0

        atexit.register(self.stop)

    def stop(self) -> None:
        """Close and unlink all shared memory buffers owned by this writer."""
        for buffer in self.buffers:
            for bufferfield in buffer.values():
                try:
                    bufferfield.stop()
                except Exception:
                    pass
        atexit.unregister(self.stop)

    def __call__(self, msg: BaseIdl) -> None:
        """
        Write data to shared memory and publish buffer number via DDS.

        Args:
            msg (BaseIdl): The data to write to shared memory.
        """
        # Rotate to the next buffer index
        self.buffer_idx = (self.buffer_idx + 1) % self.nr_of_buffers

        buffer = self.buffers[self.buffer_idx]
        # Write each field to the shared memory buffer
        for key, bufferfield in buffer.items():
            # Copy data from the message to the shared memory array
            bufferfield.shared_array[:] = getattr(msg, key)[:]

        # Publish the buffer number and timestamp via DDS
        self.buffer_nr_writer(BufferNrSample(timestamp=time.time(), nr=self.buffer_idx))

    def __make_shared_memory(self) -> List[Dict[str, SMBufferWriteField]]:
        """
        Create shared memory buffers based on the buffer template.

        Returns:
            List[Dict[str, SMBufferWriteField]]: A list of dictionaries, each containing buffer fields.
        """
        # Initialize a list to hold buffers for each buffer index
        buffers: List[Dict[str, SMBufferWriteField]] = [
            {} for _ in range(self.nr_of_buffers)
        ]

        # Iterate over each field defined in the buffer template
        for name, shape, dtype, nbytes in self.buffer_template.get_fields():
            # For each buffer index, create a shared memory field
            for buffer_idx in range(self.nr_of_buffers):
                buffers[buffer_idx][name] = SMBufferWriteField(
                    f"{self.topic_name}.{name}.buffer_{buffer_idx}",
                    shape,
                    dtype,
                    nbytes,
                )

        return buffers



"""
This module provides classes to read data from shared memory using Cyclone DDS.

Classes:
    SharedMemoryNoResourceTracker: A wrapper around SharedMemory that unregisters from the resource tracker.
    SMBufferReadField: Manages a numpy array backed by shared memory.
    SMReader: Reads shared memory buffers using DDS for synchronization.
"""

import atexit
import time
from multiprocessing import shared_memory, resource_tracker
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from cyclonedds.domain import DomainParticipant
from loguru import logger

from airo_ipc.cyclone_shm.idl.defaults.buffer_nr import BufferNrSample
from airo_ipc.cyclone_shm.idl_shared_memory.base_idl import BaseIdl
from airo_ipc.cyclone_shm.patterns.ddsreader import DDSReader


class WaitingForFirstMessageException(Exception):
    pass


class SharedMemoryNoResourceTracker:
    """
    Wrapper around shared_memory.SharedMemory that unregisters from the resource tracker.

    This prevents the resource_tracker from automatically unlinking the shared memory segment
    when the process exits, which is useful when multiple processes need to share the memory.
    """

    def __init__(self, name: str) -> None:
        """
        Initialize the shared memory segment.

        Args:
            name (str): The name of the existing shared memory block to attach to.
        """
        self.shm = shared_memory.SharedMemory(name=name)
        # Unregister from resource_tracker to prevent automatic unlinking
        resource_tracker.unregister("/" + self.shm.name, "shared_memory")

    def close(self) -> None:
        """Close the shared memory segment."""
        self.shm.close()

    @property
    def buf(self) -> Union[memoryview, None]:
        """Return the buffer interface to the shared memory."""
        return self.shm.buf


class SMBufferReadField:
    """
    Manages a numpy array backed by shared memory for reading.

    This class attaches to a shared memory segment and creates a numpy array
    that uses the shared memory as its buffer.
    """

    def __init__(self, name: str, shape: Tuple[int, ...], dtype: Any):
        """
        Initialize the shared memory buffer field.

        Args:
            name (str): The name of the shared memory block.
            shape (tuple): The shape of the numpy array.
            dtype (numpy.dtype): The data type of the numpy array.
        """
        self.shm = SharedMemoryNoResourceTracker(name=name)
        self.shared_array: np.ndarray = np.ndarray(
            shape, dtype=dtype, buffer=self.shm.buf
        )

    def stop(self) -> None:
        """Close the shared memory segment."""
        self.shm.close()


class SMReader:
    """
    Reads shared memory buffers using Cyclone DDS for synchronization.

    This class reads data from shared memory that is synchronized using DDS.
    It listens to a DDS topic for buffer numbers and reads the corresponding
    buffers from shared memory.
    """

    def __init__(
        self,
        domain_participant: DomainParticipant,
        topic_name: str,
        idl_dataclass: BaseIdl,
        nr_of_buffers: int = 3,
    ):
        """
        Initialize the shared memory reader.

        Args:
            domain_participant (DomainParticipant): The DDS domain participant.
            topic_name (str): The base name of the topic.
            idl_dataclass (BaseIdl): The template defining the buffer structure.
            nr_of_buffers (int): The number of shared memory buffers per field. Raise this value to lower the chance of data races. Lower values reduce memory usage at the risk of data races. This value should be synchronized between writers and readers.
        """
        self.domain_participant = domain_participant
        self.topic_name = topic_name
        self.buffer_template = idl_dataclass
        self.nr_of_buffers = nr_of_buffers

        # Create a DDS reader for buffer numbers
        self.buffer_nr_reader = DDSReader(
            domain_participant=domain_participant,
            topic_name=f"{topic_name}__buffer_nr",
            idl_dataclass=BufferNrSample,
        )
        # Wait for the writer to start publishing
        self.__wait_for_writer()

        # Load the shared memory buffers
        self.buffers = self.__load_shared_memory()

        atexit.register(self.stop)

    def stop(self) -> None:
        """Close all shared memory buffers owned by this reader."""
        for buffer in self.buffers:
            for bufferfield in buffer.values():
                try:
                    bufferfield.stop()
                except Exception:
                    pass
        atexit.unregister(self.stop)

    def __call__(self) -> BaseIdl:
        """
        Read the latest data from shared memory.
        No copy is made; the returned instance references the shared memory directly.
        You should copy the data immediately if you need to retain it.

        Returns:
            An instance of buffer_template.__class__ containing the data.

        Raises:
            WaitingForFirstMessageException: If no data is available yet.
        """
        buffer_nr_sample = self.buffer_nr_reader()
        if buffer_nr_sample is None:
            raise WaitingForFirstMessageException
        buffer = self.buffers[buffer_nr_sample.nr]

        kwargs = {}
        for key, bufferfield in buffer.items():
            kwargs[key] = bufferfield.shared_array

        return self.buffer_template.__class__(**kwargs)

    def read_into(self, output_instance: BaseIdl) -> None:
        """
        Read the latest data from shared memory into the provided instance.
        This avoids allocating a new instance of the buffer template class,
        which can be a minor performance improvement in very high-frequency scenarios
        and reduce pressure on the garbage collector.

        No copy is made; the returned instance references the shared memory directly.
        You should copy the data immediately if you need to retain it.

        Args:
            output_instance (BaseIdl): An instance of the buffer template class to populate.

        Raises:
            WaitingForFirstMessageException: If no data is available yet.
        """
        buffer_nr_sample = self.buffer_nr_reader()
        if buffer_nr_sample is None:
            raise WaitingForFirstMessageException
        buffer = self.buffers[buffer_nr_sample.nr]

        for key, bufferfield in buffer.items():
            setattr(output_instance, key, bufferfield.shared_array)

    def __load_shared_memory(self) -> List[Dict[str, SMBufferReadField]]:
        """
        Load shared memory buffers based on the buffer template.

        Returns:
            list: A list of dictionaries, each containing buffer fields.
        """
        # Initialize a list to hold buffers for each buffer index
        buffers: List[Dict[str, SMBufferReadField]] = [
            {} for _ in range(self.nr_of_buffers)
        ]

        # Iterate over each field defined in the buffer template
        for name, shape, dtype, nbytes in self.buffer_template.get_fields():
            # For each buffer index, create a shared memory field
            for buffer_idx in range(self.nr_of_buffers):
                buffers[buffer_idx][name] = SMBufferReadField(
                    f"{self.topic_name}.{name}.buffer_{buffer_idx}",
                    shape,
                    dtype,
                )
        return buffers

    def __wait_for_writer(
        self, timeout: Optional[float] = None, warn_every: int = 60
    ) -> None:
        """
        Wait for the DDS writer to start publishing buffer numbers.

        Args:
            timeout (float, optional): The maximum time to wait in seconds. If None, wait indefinitely.
            warn_every (int): The interval in seconds to log a warning message.

        Raises:
            RuntimeError: If the writer does not start within the timeout period.
        """
        t0 = time.time()
        warned = 0
        while self.buffer_nr_reader() is None and (
            timeout is None or (time.time() - t0) < timeout
        ):
            if warn_every * (warned + 1) < (time.time() - t0):
                warned += 1

                # Construct a warning message about waiting time
                warning_msg = (
                    f"Shared Memory Reader {self.topic_name} has been "
                    f"waiting for Shared Memory Writer for more than "
                    f"{warned * warn_every} seconds. "
                )
                if timeout is None:
                    warning_msg += "No timeout defined; will wait indefinitely."
                else:
                    warning_msg += (
                        f"Timeout: {timeout} seconds. Will wait for "
                        f"{timeout - warned * warn_every} more seconds."
                    )

                logger.warning(warning_msg)

            # Sleep briefly to yield control and prevent busy waiting
            time.sleep(0.01)

        if self.buffer_nr_reader() is None:
            # If the writer has not started within the timeout, raise an error
            raise RuntimeError(
                f"Shared Memory Reader {self.topic_name} timed out waiting for buffer."
            )



from dataclasses import dataclass, field

from cyclonedds.idl import IdlStruct


@dataclass
class BufferNrSample(IdlStruct, typename="BufferNr.Msg"):
    timestamp: float = field(metadata={"id": 0})

    nr: int = field(metadata={"id": 1})
